from django.urls import path

from . import views

app_name = "django_tiles_gl"

urlpatterns = [
    path("<int:z>/<int:x>/<int:y>.pbf", views.tile, name="tile"),
    path(
        "style/openmaptiles.json", views.openmaptiles_style, name="openmaptiles_style"
    ),
    path("tiles.json", views.tilejson, name="tilejson"),
]



import json
from importlib.metadata import metadata
from pathlib import Path

from django.conf import settings
from django.contrib.staticfiles.storage import staticfiles_storage
from django.http import HttpResponse, JsonResponse
from django.shortcuts import render
from django.urls import reverse
from django.views.decorators.cache import cache_control

from .mbtiles import MissingTileError, open_mbtiles
from .utils import center_from_bounds, build_absolute_uri

DEFAULT_ZOOM = 13
DEFAULT_MINZOOM = 7
DEFAULT_MAXZOOM = 15
WORLD_BOUNDS = [-180, -85.05112877980659, 180, 85.0511287798066]

MONTH_SECONDS = 60 * 60 * 24 * 30


@cache_control(max_age=MONTH_SECONDS)
def tile(request, z, x, y):
    with (open_mbtiles()) as mbtiles:
        try:
            data = mbtiles.tile(z, x, y)
            response = HttpResponse(
                content=data,
                status=200,
            )
            response["Content-Type"] = "application/x-protobuf"
            response["Content-Encoding"] = "gzip"

            return response

        except MissingTileError:
            return HttpResponse(
                status=204,
            )


def openmaptiles_style(request):
    if hasattr(settings, "MBTILES_CENTER"):
        center = settings.MBTILES_CENTER
    else:
        with open_mbtiles() as mbtiles:
            metadata = mbtiles.metadata()
            bounds = metadata.get("bounds", WORLD_BOUNDS)
            center = metadata.get("center", center_from_bounds(bounds, DEFAULT_ZOOM))

    base_url = staticfiles_storage.url("django-tiles-gl/")
    if not base_url.startswith("http"):
        base_url = build_absolute_uri(request, base_url)

    tilejson_url = build_absolute_uri(request, reverse("django_tiles_gl:tilejson"))

    app_path = Path(__file__).parent
    style_json_path = app_path / "templates" / "django_tiles_gl" / "style.json"
    with style_json_path.open() as f:
        style = json.load(f)

    style["center"] = [center[0], center[1]]
    style["zoom"] = center[2]
    style["sprite"] = base_url + "sprites/sprite"
    style["glyphs"] = base_url + r"fonts/{fontstack}/{range}.pbf"
    style["sources"] = {"openmaptiles": {"type": "vector", "url": tilejson_url}}

    return JsonResponse(style)


def tilejson(request):
    with open_mbtiles() as mbtiles:
        metadata = mbtiles.metadata()

        # Load valid tilejson keys from the mbtiles metadata
        valid_tilejson_keys = (
            # MUST
            "name",
            "format",
            # SHOULD
            "bounds",
            "center",
            "minzoom",
            "maxzoom",
            # MAY
            "attribution",
            "description",
            "type",
            "version",
            # UNSPECIFIED
            "scheme",
        )
        spec = {key: metadata[key] for key in valid_tilejson_keys if key in metadata}

        if spec["format"] == "pbf":
            spec["vector_layers"] = metadata["json"]["vector_layers"]
        else:
            raise NotImplementedError(
                f"Only mbtiles in pbf format are supported. Found {spec['format']}"
            )

        # Optional fields
        spec["scheme"] = spec.get("scheme", "xyz")
        spec["bounds"] = spec.get("bounds", WORLD_BOUNDS)
        spec["minzoom"] = spec.get("minzoom", DEFAULT_MINZOOM)
        spec["maxzoom"] = spec.get("maxzoom", DEFAULT_MINZOOM)
        spec["center"] = spec.get(
            "center", center_from_bounds(spec["bounds"], DEFAULT_ZOOM)
        )

        # Tile defintions
        tile_url = build_absolute_uri(
            request,
            reverse("django_tiles_gl:tile", args=(0, 0, 0))
        )
        tile_url = tile_url.replace("/0/0/0.pbf", r"/{z}/{x}/{y}.pbf")
        spec["tiles"] = [tile_url]

        # Version defintion
        spec["tilejson"] = "3.0.0"

        return JsonResponse(spec)



from django.conf import settings


def split_floats(input, sep=","):
    return [float(val.strip()) for val in input.split(sep)]

def center_from_bounds(bounds, zoom):
    return [(bounds[0] + bounds[2]) / 2.0, (bounds[1] + bounds[3]) / 2.0, zoom]

def build_absolute_uri(request, location=None):
    uri = request.build_absolute_uri(location)

    if getattr(settings, "MBTILES_FORCE_SSL", False) and uri[:7].lower() == "http://":
        uri = "https://" + uri[7:]

    return uri



import json
import os
import sqlite3
from curses import meta

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

from .utils import split_floats


class MissingTileError(Exception):
    pass


class MBTilesNotFoundError(Exception):
    pass


class MBTilesInvalid(Exception):
    pass


def open_mbtiles():
    database = getattr(settings, "MBTILES_DATABASE")
    if not database:
        raise ImproperlyConfigured("The MBTILES_DATABASE setting must not be empty.")
    return MBTiles(str(database))


class MBTiles:
    _connection = None

    def __init__(self, database):
        self._database = database

    def connect(self):
        if not os.path.exists(self._database):
            raise MBTilesNotFoundError(f"MBTiles database {self._database} does not exist")

        self._connection = sqlite3.connect(
            database=self._database,
            timeout=1.0,
            detect_types=0,
            isolation_level=None,
            check_same_thread=False,
        )

    def close(self):
        if self._connection:
            self._connection.close()
            self._connection = None

    def metadata(self):
        cursor = self._connection.cursor()
        cursor.execute("SELECT name, value FROM metadata")
        metadata = {row[0]: row[1] for row in cursor}
        cursor.close()

        self._validate_metadata(metadata)
        self._parse_metadata_bounds(metadata)
        self._parse_metadata_center(metadata)
        self._parse_metadata_zoom(metadata)
        self._parse_metadata_json(metadata)

        self._add_scheme_with_default(metadata)

        return metadata

    def _validate_metadata(self, metadata):
        required = ["name", "format"]
        for field in required:
            if field not in metadata:
                raise MBTilesInvalid(f'Missing required metadata field "{field}"')

        if metadata["format"] == "pbf":
            if "json" not in metadata:
                raise MBTilesInvalid(f'Missing required metadata field "json"')

    def _parse_metadata_bounds(self, metadata):
        if "bounds" in metadata:
            metadata["bounds"] = split_floats(metadata["bounds"])

    def _parse_metadata_center(self, metadata):
        if "center" in metadata:
            metadata["center"] = split_floats(metadata["center"])

    def _parse_metadata_zoom(self, metadata):
        for zoom in ("minzoom", "maxzoom"):
            if zoom in metadata:
                metadata[zoom] = float(metadata[zoom])

    def _parse_metadata_json(self, metadata):
        if "json" in metadata:
            metadata["json"] = json.loads(metadata["json"])

    # FIXME: WHAT SHOULD BE THE DEFAULT?
    def _add_scheme_with_default(self, metadata, default="tms"):
        metadata["scheme"] = metadata.get("scheme", default)

    # def _flip_y(self, y, z):
    #     return 2**z - 1 - y

    def tile(self, z, x, y):
        cursor = self._connection.cursor()
        # tms_y = self._flip_y(y, z)
        cursor.execute(
            "SELECT tile_data FROM tiles WHERE zoom_level=? AND tile_column=? AND tile_row=?;",
            # (z, x, tms_y),
            (z, x, y),
        )
        tile = cursor.fetchone()
        cursor.close()

        if not tile:
            raise MissingTileError()
        return tile

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.close()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['django_tiles_gl', 'django_tiles_gl.templatetags']

package_data = \
{'': ['*'],
 'django_tiles_gl': ['static/django-tiles-gl/fonts/Open Sans Bold Italic/*',
                     'static/django-tiles-gl/fonts/Open Sans Bold/*',
                     'static/django-tiles-gl/fonts/Open Sans Extra Bold '
                     'Italic/*',
                     'static/django-tiles-gl/fonts/Open Sans Extra Bold/*',
                     'static/django-tiles-gl/fonts/Open Sans Italic/*',
                     'static/django-tiles-gl/fonts/Open Sans Light Italic/*',
                     'static/django-tiles-gl/fonts/Open Sans Light/*',
                     'static/django-tiles-gl/fonts/Open Sans Regular/*',
                     'static/django-tiles-gl/fonts/Open Sans Semibold Italic/*',
                     'static/django-tiles-gl/fonts/Open Sans Semibold/*',
                     'static/django-tiles-gl/maplibre/*',
                     'static/django-tiles-gl/sprites/*',
                     'templates/django_tiles_gl/*']}

install_requires = \
['Django>=3.1,<4.1']

setup_kwargs = {
    'name': 'django-tiles-gl',
    'version': '0.3.6',
    'description': 'Integrated Django Vector Tile Server based on mbtiles',
    'long_description': '# Django Tiles GL\n\nIntegrated Django Vector Tile Server based on mbtiles.\n\n\n## Description\n\nSimple app to serve [Mabpox Vector Tiles](https://docs.mapbox.com/data/tilesets/guides/vector-tiles-standards/) directly from [MBTiles files](https://github.com/mapbox/mbtiles-spec) via Django views.\n\nDjango Tiles has a minimal dependencies. It does not require GeoDjano or any other libraries. Its only dependency is Django itself.\n\nDjango Tiles GL **does not create raster tiles**. It may only be used with map libraries that support to render vector tiles like  [MapLibre](https://maplibre.org/) or [OpenLayers](https://openlayers.org/).\n\nDjango Tiles GL contains the [OSM Bright map style](https://openmaptiles.org/styles/#osm-bright) which can be used to render Vector Tiles following the [OpenMapTiles vector tile  schema](https://openmaptiles.org/schema).\n\nNote that this default style is using [OpenSans fonts](https://github.com/openmaptiles/fonts) which does only contain Latin, Greek and Cyrillic alphabets.\n\nOther tile schemes are possible by creating a custom [map style specification](https://docs.mapbox.com/mapbox-gl-js/style-spec/) and referencing Django Tiles GL [TileJSON endpoint](https://github.com/mapbox/tilejson-spec) as a source.\n\n## Usage\n\nSee the [`demo`](https://github.com/kleingeist/django-tiles-gl/tree/main/demo) Django application for a simple usage example.\n\n\n### Setup\n\n- Add `django_tiles_gl` to you `INSTALLED_APPS` setting.\n- Add `django_tiles_gl.urls` to your url patterns.\n  For example with the `tiles` prefix:\n  ```python\n  urlpatterns = [\n      ...\n      path("tiles/", include("django_tiles_gl.urls")),\n  ]\n  ```\n- Set path to your MBTiles files in you application settings.\n  ```python\n  MBTILES_DATABASE = BASE_DIR / "demo" / "data" / "berlin.mbtiles"\n  ```\n- Optionally set the default center to be set on the default map style.\n  ```python\n  MBTILES_CENTER = [13.4, 52.5, 13]   # [longitude, latitude, zoom]\n  ```\n- Optionally force absolute urls to use SSL by prefixing them with "https://".\n  This might be required if you app is running behind a reverse proxy and you\n  are not able to set [`SECURE_PROXY_SSL_HEADER`](https://docs.djangoproject.com/en/4.0/ref/settings/#secure-proxy-ssl-header)\n  from the SSL enabled proxy server.\n  ```python\n  MBTILES_FORCE_SSL = True\n  ```\n\n### Views\nTo render a map you have to include a JavaScript mapping library and refer to the `tile` endpoint or the default integrated style.\n\nDjango Tiles GL provides the following endpoints:\n\n- `{% url \'django_tiles_gl:openmaptiles_style\' %}` - Default [OpenMapTiles style defintion](https://openmaptiles.org) using the [OSM Bright map style](https://openmaptiles.org/styles/#osm-bright).\n- `{% url \'django_tiles_gl:tilejson\' %}` - [TileJSON](https://github.com/mapbox/tilejson-spec) describing the configured MBTiles files and providing the correct tile urls.\n- `{% url \'django_tiles_gl:tile\' x y z %}` - Actual tile endpoint, returning vector data in the [PBF format](https://wiki.openstreetmap.org/wiki/PBF_Format).\n\n\nDjango Tiles GL is bundles with [MapLibre](https://maplibre.org/) and provides a template tag for easy inclusion. A minimal working example has to contain the following defintions:\n\n```html\n{% load tiles_gl_tags %}\n<!DOCTYPE html>\n<html>\n<head>\n    {% maplibre_head %}\n\n    <style>\n        body { margin:0; padding:0; }\n        #map { position:absolute; top:0; bottom:0; width:100%; }\n    </style>\n</head>\n<body>\n\n<div id=\'map\'></div>\n\n<script>\nvar map = new maplibregl.Map({\n\tcontainer: \'map\',\n\tstyle: \'{% url \'django_tiles_gl:openmaptiles_style\' %}\',\n});\n</script>\n\n</body>\n</html>\n```\n\n## Data / MBTiles generation\n\nThere are mutiple tools to generate valid MBTiles databases. The easiest to use with Django Tiles GL is [OpenMapTiles](https://github.com/openmaptiles/openmaptiles) as it is compatible with the bundled default style.\n\nFor a quickstart you may generate the MbTiles for an area with the following commands:\n```sh\ngit clone https://github.com/openmaptiles/openmaptiles.git\ncd openmaptiles\n./quickstart.sh <area>\n```\n\nFo further information and optiones see [https://github.com/openmaptiles/openmaptiles](https://github.com/openmaptiles/openmaptiles)\n\n\n\n## Further Topics\n\n### Caching\n\nIt is advised setup a caching proxy for the `tile` endpoint. Please refer to you HTTP servers documentation. For example the [nginx proxy cache config](https://nginx.org/en/docs/http/ngx_http_proxy_module.html#proxy_cache)\n\n\n\n## Other Django tile server solutions\n\n- [django-geojson-tiles](https://github.com/glenrobertson/django-geojson-tiles) - Generates GeoJSON tiles from a GeoDjango model. No MBTiles support.  Not to be used as a base layer.\n\n- [django-vectortiles](https://github.com/submarcos/django-vectortiles) - Generates Vector Tile layers from GeoDjango. No MBTiles support. Not to be used as a base layer.\n\n- [django-mbtiles](https://github.com/makinacorpus/django-mbtiles) - Uses MBTiles to generate rastered tiles and [UTFGrid](https://github.com/mapbox/mbtiles-spec/blob/master/1.1/utfgrid.md). Does not support modern vector tiles. Strong inspiration for this project.\n',
    'author': 'Johannes Dillmann',
    'author_email': 'dev@ae35.de',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/kleingeist/django-tiles-gl',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
